function [U_ii_new, U_jj_new] = Grassmann_preconditioned_update(ii, jj, U_cells, problem, options)
% Take a gradient step by preconditioning the gradients.
% U_ii ---> U_ii_new
% U_jj ---> U_jj_new

% Part of the codes for Riemannian gossip algorithms for 
% decentralized matrix completion.
% Original authors: Bamdev Mishra <bamdevm@gmail.com> and
%                   Hiroyuki Kasai <kasai@is.uec.ac.jp>.
    
    % Problem information
    m = problem.m;
    r = problem.r;
    N = problem.N;
    
    % Options
    rho = options.rho;
    stepsize = options.stepsize;
    lambda = options.lambda;
    
    % Define the Grassmann manifold
    problem.M = grassmannfactory(m, r);
    
    
    % Compute the factors at ii and jj nodes.
    U_ii = U_cells{ii};
    W_ii = mylsqfit(U_ii, ii, problem, options); % Compute W by solving a least-square problem.
    
    U_jj = U_cells{jj};
    W_jj = mylsqfit(U_jj, jj, problem, options);
    
    
    % Balancing the probabilities.
    alpha_ii = 0.5;
    alpha_jj = 0.5;
    
    if ii == 1
        alpha_ii = 1;
    end
    
    if ii == N - 1
        alpha_jj = 1;
    end
    
   
    % Computing ing gradients at the nodes.
    % It contains a weighted sum of objective function plus sharing.
    
    UW_ii = spmaskmult(U_ii, W_ii', problem.I_cells{ii}, problem.J_cells{ii});
    RU_ii = (1- lambda^2)*(UW_ii - problem.entries_cells{ii})  - (lambda^2)* problem.entries_cells{ii};
    grad_ii = alpha_ii*multsparsefull(RU_ii, W_ii, problem.masks_cells{ii}) ...
        + (alpha_ii*lambda^2)*U_ii*(W_ii'*W_ii) ...
        - rho*problem.M.log(U_ii, U_jj);
    
    UW_jj = spmaskmult(U_jj, W_jj', problem.I_cells{jj}, problem.J_cells{jj});
    RU_jj = (1- lambda^2)*(UW_jj - problem.entries_cells{jj}) -  - (lambda^2)* problem.entries_cells{jj};
    grad_jj = alpha_jj*multsparsefull(RU_jj, W_jj, problem.masks_cells{jj}) ...
        + (alpha_jj*lambda^2)*U_jj*(W_jj'*W_jj) ...
        - rho*problem.M.log(U_jj, U_ii);
    
    
    % Precondition the gradients appropriately by a positive
    % combination of local and global curvature information.
    scale_ii = alpha_ii*(W_ii'*W_ii) + rho*eye(r);
    scale_jj = alpha_jj*(W_jj'*W_jj) + rho*eye(r);
    
    
    scaled_grad_ii = grad_ii/scale_ii;
    scaled_grad_jj = grad_jj/scale_jj;
    
    
    % Take a gradient step at both ii and jj nodes.    
    U_ii_new = problem.M.exp(U_ii, -scaled_grad_ii, stepsize);
    U_jj_new = problem.M.exp(U_jj, -scaled_grad_jj, stepsize);
        
end


