function t = mystepsizeguess(U_cells, problem, options)
%
% Linearization of the cost function to compute a stepsize guess for each
% agent. 
%

% Part of the codes for Riemannian gossip algorithms for 
% decentralized matrix completion.
% Original authors: Bamdev Mishra <bamdevm@gmail.com> and
%                   Hiroyuki Kasai <kasai@is.uec.ac.jp>.
     
    r = problem.r;
    N = problem.N;
    
    ii = randi(problem.N -1);
    jj = ii + 1;
    
    U_ii = U_cells{ii};
    W_ii = mylsqfit(U_ii, ii, problem, options); % Compute W by solving a least-square problem.
    
    U_jj = U_cells{jj};
    W_jj = mylsqfit(U_jj, jj, problem, options);
    
    % Acols
    masktemp_ii = problem.masks_cells{ii};
    setsparseentries(masktemp_ii, ones(size(problem.entries_cells{ii})));
    Achol_ii = spbuildmatrix(masktemp_ii, U_ii, eps);
    
    masktemp_jj = problem.masks_cells{jj};
    setsparseentries(masktemp_jj, ones(size(problem.entries_cells{jj})));
    Achol_jj = spbuildmatrix(masktemp_jj, U_jj, eps);
    
   
     % Balancing the probabilities.
    alpha_ii = 0.5;
    alpha_jj = 0.5;
    
    if ii == 1
        alpha_ii = 1;
    end
    
    if ii == N - 1
        alpha_jj = 1;
    end
    
    % Computing ing gradients at the nodes.
    % It contains a weighted sum of objective function plus sharing.
    UW_ii = spmaskmult(U_ii, W_ii', problem.I_cells{ii}, problem.J_cells{ii});
    RU_ii = UW_ii - problem.entries_cells{ii};
    grad_ii = alpha_ii*multsparsefull(RU_ii, W_ii, problem.masks_cells{ii}) ...
        - options.rho*problem.M.log(U_ii, U_jj);
    
    UW_jj = spmaskmult(U_jj, W_jj', problem.I_cells{jj}, problem.J_cells{jj});
    RU_jj = UW_jj - problem.entries_cells{jj};
    grad_jj = alpha_jj*multsparsefull(RU_jj, W_jj, problem.masks_cells{jj}) ...
        - options.rho*problem.M.log(U_jj, U_ii);
    
    
    % Type of update
    h0 = options.update;
    h1 = @Grassmann_update;
    h2 = @Grassmann_preconditioned_update;
    h3 = @Euclidean_update;

    if isequal(h0, h1) || isequal(h0, h3)
        H_ii = -grad_ii;
        H_jj = -grad_jj;
        
        
    elseif isequal(h0, h2)
        H_ii = -grad_ii/(alpha_ii*(W_ii'*W_ii) + options.rho*eye(r));
        H_jj = -grad_jj/(alpha_jj*(W_jj'*W_jj) + options.rho*eye(r));
        
    else
       error(['Unknown options.update. ' ...
                    'Should be Grassmann_update or Grassmann_preconditioned_update.']);
    end
    
    
    
    % Compute derivative of W along H
    
    HW_ii = spmaskmult(H_ii, W_ii', problem.I_cells{ii}, problem.J_cells{ii});   
    Q1_ii = multfullsparse(H_ii.', RU_ii, masktemp_ii);
    Q2_ii = multfullsparse(U_ii.', HW_ii, masktemp_ii);
    dW_ii = -(blocksolve(Achol_ii, Q1_ii + Q2_ii))';
    
    HW_jj = spmaskmult(H_jj, W_jj', problem.I_cells{jj}, problem.J_cells{jj});   
    Q1_jj = multfullsparse(H_jj.', RU_jj, masktemp_jj);
    Q2_jj = multfullsparse(U_jj.', HW_jj, masktemp_jj);
    dW_jj = -(blocksolve(Achol_jj, Q1_jj + Q2_jj))';
    
    
    UdW_ii = spmaskmult(U_ii, dW_ii', problem.I_cells{ii}, problem.J_cells{ii}) ...
             + spmaskmult(H_ii, W_ii', problem.I_cells{ii}, problem.J_cells{ii});   
    UdW_jj = spmaskmult(U_jj, dW_jj', problem.I_cells{jj}, problem.J_cells{jj}) ...
             + spmaskmult(H_jj, W_jj', problem.I_cells{jj}, problem.J_cells{jj});
         
    num = UdW_ii'*RU_ii + UdW_jj'*RU_jj;
    den = UdW_ii'*UdW_ii + UdW_jj'*UdW_jj;
    
    Hjj_transported = problem.M.transp(U_ii, U_jj, H_jj);
    num = num - options.rho*trace(problem.M.log(U_ii, U_jj)'*(H_ii - Hjj_transported));
    den = den + options.rho*norm(H_ii - Hjj_transported , 'fro')^2;
        
    t = -num/den;
end



function X = blocksolve(R, B)
    % R is a cell containing n upper triangular square matrices of size r.
    % B is a matrix of size r x n. X is a matrix of size r x n where the i-th
    % column is the solution of the equation A{i}x = B(:, i), with
    % A{i} = R{i}.'*R{i}. This is: the R{i} matrices are the upper triangular
    % Cholesky factors of the matrices A{i} (which we do not require/store).
    
    X = cholsolvecell(R, B);
    
end