function [U_cells, infos, options] = online_decentralized_MC(U_cells0, problem, options)
% Online implementation of the gossip algorithm.
% 
% options.update = @Grassmann_update and @Grassmann_preconditioned_update.
%                  Types of gradient updates.
% options.lambda = Regularization parameter. Equal to the one in RTRMC.
% options.stepsize = Initial stepsize.
% options.iters = Number of iterations.
% options.rho = Weight parameter that weighs completion and consensus
%               terms.
% options.computeNMAE = To compute the NMAE scores if there is test data.
%
% B. Mishra, H. Kasai, P. Jawanpuria, and A. Saroop
% "A Riemannian gossip approach to decentralized subspace learning on Grassmann manifold,"
% Technical report, arXiv preprint arXiv:1705.00467, 2017.
%
%
% Part of the codes for Riemannian gossip algorithms for 
% decentralized matrix completion.
% Original authors: Bamdev Mishra <bamdevm@gmail.com> and
%                   Hiroyuki Kasai <kasai@is.uec.ac.jp>.
 
    
    % Mandatory problem information
    m = problem.m;
    r = problem.r;
    N = problem.N;
   

    % Define the Grassmann manifold
    problem.M = grassmannfactory(m, r);
    
    % Initialization of U at all the N nodes.
    if isempty(U_cells0)
        for ii = 1:N
            U_cells0{ii} = problem.M.rand();
        end
    end
    U_cells = U_cells0;
    
    % Options
    if ~isfield(options,'update'); options.update = @Grassmann_update; end % Type of update
    if ~isfield(options, 'lambda'); options.lambda = 1e-5; end % Regularization parameter.
    if ~isfield(options,'stepsize'); options.stepsize = 100*(N-1)*mystepsizeguess(U_cells0, problem, options); end % Initial stepsize.
    if ~isfield(options, 'stepsize_decay'); options.stepsize_decay = 1/options.stepsize; end % Stepsize decay parameter.
    if ~isfield(options,'iters'); options.iters = (N-1)*100; end % 100 updates of nodes 1 and N and 200 updates of the rest.
    if ~isfield(options, 'rho'); options.rho = 1e3; end % Weight parameter.
    if ~isfield(options, 'computeNMAE'); options.computeNMAE = false; end % Compute NMAE.
    if ~isfield(options, 'verbosity'); options.verbosity = true; end % Verbose.


    stepsize0 = options.stepsize;
    lambda = options.lambda;

    % Infos
    entries_max_min = -inf;
    for ii = 1 : N
        infos{ii}.cost = [];
        infos{ii}.commcost=[];
        infos{ii}.time = 0;
        infos{ii}.nentries = problem.nentries_cells{ii};
        
        if ~isempty(problem.entries_test_cells)
            infos{ii}.cost_test = [];  
            infos{ii}.nentries_test = problem.nentries_test_cells{ii};
            
            if options.computeNMAE
                infos{ii}.nmae_test = [];
                entries_max_min = max(entries_max_min, max(problem.entries_cells{ii}) - min(problem.entries_cells{ii}));
            end
        end 
    end
    
    % The N+1 structure is used for storing other information.
    infos{N+1}.time = 0;
    infos{N+1}.iter = [];
    
    
    % Stepsize sequence rate parameters
    a = stepsize0;
    b = stepsize0*options.stepsize_decay;
    
    % Main algorithm.
    % We assume that the nodes are labeled according to their proximity
    % as follows:
    % for i <= N - 1 the nodes i and i + 1 are neighbors.
    % We pick a node i (< N) uniformly at random.
    
    for kk = 1 : options.iters
        
        ii = randi(N-1); % Chosen node from 1 to N-1 with uniform probability.
        jj = ii + 1; % Its neighbor.
        
        % Stepsize is decreased at every iteration.
        options.stepsize = a/(1 + b*kk);
        
        % Update the nodes with a gradient step.
        tic;
        [U_ii, U_jj] = options.update(ii, jj, U_cells, problem, options);
        timeelapsed = toc;
        
        % Update U_cells
        U_cells{ii} = U_ii;
        U_cells{jj} = U_jj;
        
        % Collect stats: time
        infos{ii}.time = [infos{ii}.time; infos{ii}.time(end) + timeelapsed];
        infos{jj}.time = [infos{jj}.time; infos{jj}.time(end) + timeelapsed];
        
        % Collect stats per iteration
        infos{N +1}.time = [infos{N +1}.time; infos{N +1}.time(end) + timeelapsed];
        infos{N +1}.iter = [infos{N +1}.iter; kk];
        
        
        % Collect stats: cost
        W_ii = mylsqfit(U_ii, ii, problem, options);
        W_jj = mylsqfit(U_jj, jj, problem, options);
        
        UW_ii = spmaskmult(U_ii, W_ii', problem.I_cells{ii}, problem.J_cells{ii});
        RU_ii = UW_ii - problem.entries_cells{ii};
        cost1 =  ((RU_ii'*RU_ii) + lambda.^2*sum(W_ii(:).^2) - lambda.^2*sum(UW_ii.^2))/problem.nentries_cells{ii}; % BM: modifies for regularization.
        
        UW_jj = spmaskmult(U_jj, W_jj', problem.I_cells{jj}, problem.J_cells{jj});
        RU_jj = UW_jj - problem.entries_cells{jj};
        cost2 =  ((RU_jj'*RU_jj) + lambda.^2*sum(W_jj(:).^2) - lambda.^2*sum(UW_jj.^2))/problem.nentries_cells{jj};  % BM: modifies for regularization.
        
        commcost12 = problem.M.norm(U_ii, problem.M.log(U_ii, U_jj))^2;

        
        % Update stats
        infos{ii}.cost = [infos{ii}.cost; cost1];
        infos{ii}.commcost = [infos{ii}.commcost; commcost12];
        infos{jj}.cost = [infos{ii+1}.cost; cost2];
        
        % Print
        if options.verbosity
            fprintf('Iteration %0.3d, cost node %d: %7.3e, cost node %d: %7.3e, communication cost: %7.3e, Stepsize %7.3e\n', kk, ii, cost1, jj, cost2, commcost12, options.stepsize);
        end
        
        
        % Collect stats: test
        if ~isempty(problem.entries_test_cells)
            UW_ii_test = spmaskmult(U_ii, W_ii', problem.I_test_cells{ii}, problem.J_test_cells{ii});
            RU_ii_test = UW_ii_test - problem.entries_test_cells{ii};
            cost1_test =  (RU_ii_test'*RU_ii_test)/problem.nentries_test_cells{ii};
            
            UW_jj_test = spmaskmult(U_jj, W_jj', problem.I_test_cells{jj}, problem.J_test_cells{jj});
            RU_jj_test = UW_jj_test - problem.entries_test_cells{jj};
            cost2_test =  (RU_jj_test'*RU_jj_test)/problem.nentries_test_cells{jj};
            
            infos{ii}.cost_test = [infos{ii}.cost_test; cost1_test];
            infos{jj}.cost_test = [infos{jj}.cost_test; cost2_test];
            
             if options.computeNMAE
                 nmae1_test =  sum(abs(RU_ii_test))/problem.nentries_test_cells{ii};
                 nmae2_test =  sum(abs(RU_jj_test))/problem.nentries_test_cells{jj};
                 nmae1_test = nmae1_test/entries_max_min;
                 nmae2_test = nmae2_test/entries_max_min;
                 
                 infos{ii}.nmae_test = [infos{ii}.nmae_test; nmae1_test];
                 infos{jj}.nmae_test = [infos{jj}.nmae_test; nmae2_test];
             end
        end
        
        
        %         % Debug: old gradient computations
        %         PA_ii = values_cells{ii}; % ii node
        %         P_ii = indicator_cells{ii};
        %         PA_jj =  values_cells{jj};% jj = ii + 1 node
        %         P_jj = indicator_cells{jj};
        %
        %
        %         grad_ii1 = alpha_ii*(P_ii.*(U_ii*W_ii') - PA_ii)*W_ii...
        %             - rho*problem.M.log(U_ii, U_jj);
        %
        %         grad_jj1 = alpha_jj*(P_jj.*(U_jj*W_jj') - PA_jj)*W_jj...
        %             - rho*problem.M.log(U_jj, U_ii);
        %         norm(grad_ii1 - grad_ii, 'fro')
        %         norm(grad_jj1 - grad_jj, 'fro')
        %
        %
        %
        %         % Debug: old cost computation
        %         cost11 = norm((P_ii.*(U_ii*W_ii') - PA_ii), 'fro')^2;
        %         cost21 = norm((P_jj.*(U_jj*W_jj') - PA_jj), 'fro')^2;
        %
        %         cost1 - cost11
        %         cost2 - cost21
        
    end
    
    % Make the length of time and cost vectors the same. The difference was
    % due to zero initialization of the time vectors.
    for ii = 1 : N
        infos{ii}.time = infos{ii}.time(2:end);
    end
    infos{N+1}.time = infos{N+1}.time(2:end);
    options.stepsize = stepsize0; % Reset, it was earlier changed.
    
end


