function [U_cells, infos, options] = parallel_decentralized_MC(U_cells0, problem, options)
% Parallel implementation of the gossip algorithm.
% 
% options.update = @Grassmann_update and @Grassmann_preconditioned_update.
%                  Types of gradient updates.
% options.lambda = Regularization parameter. Equal to the one in RTRMC.
% options.stepsize = Initial stepsize.
% options.iters = Number of iterations.
% options.rho = Weight parameter that weighs completion and consensus
%               terms.
% options.computeNMAE = To compute the NMAE scores if there is test data.
%
% B. Mishra, H. Kasai, P. Jawanpuria, and A. Saroop
% "A Riemannian gossip approach to decentralized subspace learning on Grassmann manifold,"
% Technical report, arXiv preprint arXiv:1705.00467, 2017.


% Part of the codes for Riemannian gossip algorithms for 
% decentralized matrix completion.
% Original authors: Bamdev Mishra <bamdevm@gmail.com> and
%                   Hiroyuki Kasai <kasai@is.uec.ac.jp>.
        
    % Mandatory problem information
    m = problem.m;
    r = problem.r;
    N = problem.N;
    
    % Define the Grassmann manifold
    problem.M = grassmannfactory(m, r);
    
    
    % Initialization of U at all the N nodes
    if isempty(U_cells0)
        for ii = 1:N
            U_cells0{ii} = problem.M.rand();
        end
    end
    U_cells = U_cells0;
    
    % Options
    if ~isfield(options,'update'); options.update = @Grassmann_update; end % Type of update
    if ~isfield(options, 'lambda'); options.lambda = 1e-5; end % Regularization parameter.
    if ~isfield(options,'stepsize'); options.stepsize = 200*mystepsizeguess(U_cells0, problem, options); end % Initial stepsize
    if ~isfield(options, 'stepsize_decay'); options.stepsize_decay = 1/options.stepsize; end % Stepsize decay parameter.
    if ~isfield(options,'iters'); options.iters = 200; end; % 100 updates of nodes 1 and N and 200 updates of the rest.
    if ~isfield(options, 'rho'); options.rho = 1e3; end % Weight parameter
    if ~isfield(options, 'computeNMAE'); options.computeNMAE = false; end % Compute NMAE.
    if ~isfield(options, 'verbosity'); options.verbosity = true; end % Verbose.


    stepsize0 = options.stepsize;
    lambda = options.lambda;
    
    % Intialization for collecting the statistics
    entries_max_min = -inf;
    for ii = 1 : N
        infos{ii}.cost = [];
        infos{ii}.commcost=[];
        infos{ii}.time = 0;
        
        if ~isempty(problem.entries_test_cells)
            infos{ii}.cost_test = [];
            if options.computeNMAE
                infos{ii}.nmae_test = [];
                entries_max_min = max(entries_max_min, max(problem.entries_cells{ii}) - min(problem.entries_cells{ii}));
            end
        end
    end
    
    % The N+1 structure is used for storing other information.
    infos{N+1}.time = 0;
    infos{N+1}.iter = [];
    
    
    round1 = 2:2:N;
    round2 = 3:2:N;
    
    
    % Main algorithm.
    % We assume that the nodes are labeled according to their proximity
    % as follows:
    % for i <= N - 1 the nodes i and i + 1 are neighbors.
    % We pick a node i (<= N-1) uniformly at random.
    %
    % In the parallel mode, round1 denotes picking 1, 3, .. nodes in
    % paralle. In round2, we pick 2, 4, ... in parallel.
    
    
    if exist('parpool', 'file') > 0
        poolobj = gcp('nocreate'); % If no pool, do not create new one.
        
        if isempty(poolobj)
            parpool('local');
        end
        
    else
        if  matlabpool('size') == 0
            matlabpool('open'); % Old matlab
        end
    end
    
    % Stepsize sequence rate parameters
    a = stepsize0;
    b = stepsize0*options.stepsize_decay;
    
    
    for kk = 1 : options.iters % Each iteration picks a set of nodes to update them in parallel.
       
        % Stepsize is decreased at every iteration.
        options.stepsize = a/(1 + b*kk);
        
        % Select the round to update.
        current_round = round1; %
        if randi(2) == 2
            % Round 2
            current_round = round2;
        end
        
        % The parallel loop for updating the nodes with a gradient step.
        Uii_current_round = cell(length(current_round), 1);
        Ujj_current_round = cell(length(current_round), 1);
        timeelapsed = Par(length(current_round)); % time storage.
        parfor ll = 1:length(current_round)
            Par.tic; 
            [Uii_current_round{ll}, Ujj_current_round{ll}] = options.update(current_round(ll)-1, current_round(ll), U_cells, problem, options);
            timeelapsed(ll) = Par.toc; 
        end
        stop(timeelapsed); % Stop par time.
        
        maxtimeelapsed = max([timeelapsed.ItStop] - [timeelapsed.ItStart]); % Time limiting worker.
        
        % Update U_cells
        for ll = 1:length(current_round)
            U_cells{current_round(ll) - 1} = Uii_current_round{ll};
            U_cells{current_round(ll)} = Ujj_current_round{ll};
        end
        
        % Collect stats: time
        for ll = 1 : length(current_round)
            ii = current_round(ll) - 1;
            jj = ii + 1;
            infos{ii}.time = [infos{ii}.time; infos{ii}.time(end) + maxtimeelapsed];
            infos{jj}.time = [infos{jj}.time; infos{jj}.time(end) + maxtimeelapsed];
        end
        
        % Collect stats per iteration
        infos{N +1}.time = [infos{N +1}.time; infos{N +1}.time(end) + maxtimeelapsed];
        infos{N +1}.iter = [infos{N +1}.iter; kk];
        
        
        % Compute W_cells
        W_cells = cell(N,1);
        for ll = 1 : length(current_round)
            ii = current_round(ll) - 1;
            jj = ii + 1;
            U_ii = U_cells{ii};
            U_jj = U_cells{jj};
            W_cells{ii} = mylsqfit(U_ii, ii, problem, options);
            W_cells{jj} = mylsqfit(U_jj, jj, problem, options);
        end
        
        % Collect stats: cost
        for ll = 1 : length(current_round)
            ii = current_round(ll) - 1;
            jj = ii + 1;
            
            U_ii = U_cells{ii};
            U_jj = U_cells{jj};
            
            W_ii = W_cells{ii};
            W_jj = W_cells{jj};
            
            UW_ii = spmaskmult(U_ii, W_ii', problem.I_cells{ii}, problem.J_cells{ii});
            RU_ii = UW_ii - problem.entries_cells{ii};
            cost1 =  ((RU_ii'*RU_ii) + lambda.^2*sum(W_ii(:).^2) - lambda.^2*sum(UW_ii.^2))/problem.nentries_cells{ii}; % BM
            
            
            UW_jj = spmaskmult(U_jj, W_jj', problem.I_cells{jj}, problem.J_cells{jj});
            RU_jj = UW_jj - problem.entries_cells{jj};
            cost2 =  ((RU_jj'*RU_jj) + lambda.^2*sum(W_jj(:).^2) - lambda.^2*sum(UW_jj.^2))/problem.nentries_cells{jj};
            
            commcost12 = problem.M.norm(U_ii, problem.M.log(U_ii, U_jj))^2;
            
            
            % Update stats
            infos{ii}.cost = [infos{ii}.cost; cost1];
            infos{ii}.commcost = [infos{ii}.commcost; commcost12];
            infos{jj}.cost = [infos{jj}.cost; cost2];
        end
        
        
        % Print
        if options.verbosity
            ii = max(current_round(1) - 1, 1);
            jj = ii + 1;
            fprintf('Iteration %0.3d, cost node %d: %7.3e, cost node %d: %7.3e, communication cost: %7.3e, Stepsize %7.3e\n', kk, ii, infos{ii}.cost(end), jj, infos{jj}.cost(end), infos{ii}.commcost(end), options.stepsize);
        end
        
        % Collect stats: test
        if ~isempty(problem.entries_test_cells)
            for ll = 1:length(current_round)
                
                ii = current_round(ll) - 1;
                jj = ii + 1;
                
                U_ii = U_cells{ii};
                U_jj = U_cells{jj};
                
                W_ii = W_cells{ii};
                W_jj = W_cells{jj};
                
                
                UW_ii_test = spmaskmult(U_ii, W_ii', problem.I_test_cells{ii}, problem.J_test_cells{ii});
                RU_ii_test = UW_ii_test - problem.entries_test_cells{ii};
                cost1_test =  (RU_ii_test'*RU_ii_test)/problem.nentries_test_cells{ii};
                
                UW_jj_test = spmaskmult(U_jj, W_jj', problem.I_test_cells{jj}, problem.J_test_cells{jj});
                RU_jj_test = UW_jj_test - problem.entries_test_cells{jj};
                cost2_test =  (RU_jj_test'*RU_jj_test)/problem.nentries_test_cells{jj};
                
                infos{ii}.cost_test = [infos{ii}.cost_test; cost1_test];
                infos{jj}.cost_test = [infos{jj}.cost_test; cost2_test];
                
                if options.computeNMAE
                    nmae1_test =  sum(abs(RU_ii_test))/problem.nentries_test_cells{ii};
                    nmae2_test =  sum(abs(RU_jj_test))/problem.nentries_test_cells{jj};
                    nmae1_test = nmae1_test/entries_max_min;
                    nmae2_test = nmae2_test/entries_max_min;
                    
                    infos{ii}.nmae_test = [infos{ii}.nmae_test; nmae1_test];
                    infos{jj}.nmae_test = [infos{jj}.nmae_test; nmae2_test];
                end
                
            end
            
        end
        
    end
    
    % Make the length of time and cost vectors the same. The difference was
    % due to zero initialization of the time vectors.
    for ii = 1 : N
        infos{ii}.time = infos{ii}.time(2:end);
    end
    infos{N+1}.time = infos{N+1}.time(2:end);
    options.stepsize = stepsize0; % Reset, it was earlier changed.
end


