clear; close all; clc;

N = 6; % Number of agents
rho = 1e3; % Regularization parameter
noiseFac = 1e-6; % Noise factor

CN = 0; % 0 for well-conditioned

m = 1000;
n = 10000;
truerank = 5;
OS = 6;
r = truerank; 
iters = 1000;

%% Data generation

M = OS*truerank*(m + n -truerank); % total entries

% The left and right factors which make up our true data matrix Y.
YL = randn(m, truerank);
YR = randn(n, truerank);

% Condition number
if CN > 0
    YLQ = orth(YL);
    YRQ = orth(YR);
        
    s1 = 1000;
    %     step = 1000; S0 = diag([s1:step:s1+(truerank-1)*step]*1); % Linear decay
    S0 = s1*diag(logspace(-log10(CN),0,truerank)); % Exponential decay
    
    YL = YLQ*S0;
    YR = YRQ;
    
    fprintf('Creating a matrix with singular values...\n')
    for kk = 1: length(diag(S0));
        fprintf('%s \n', num2str(S0(kk, kk), '%10.5e') );
    end
    
end



% Select a random set of M entries of Y.
idx = unique(ceil(m*n*rand(1,(10*M))));
idx = idx(randperm(length(idx)));

[I, J] = ind2sub([m, n],idx(1:M));
[J, inxs] = sort(J); I=I(inxs)';

% Values of Y at the locations indexed by I and J.
S = sum(YL(I,:).*YR(J,:), 2);
S_noiseFree = S;

% Add noise.
noise = noiseFac*max(S)*randn(size(S));
S = S + noise;


PA = sparse(I,J,S,m ,n);
P = sparse(I,J,1,m,n);




%% Agents specific data
l = ceil(n/N);
for ii = 1 : N
    values_cells{ii} = PA(:, 1 + (ii-1)*l: min(ii*l, n));
    indicator_cells{ii} = P(:, 1 + (ii-1)*l: min(ii*l, n));
end


% Initialization of U at all the N agents
problem.M = grassmannfactory(m, r); %  Define the Grassmann manifold
for ii = 1 : N
    U_cells0{ii} = problem.M.rand();
end



% problem structure definition
problem.m = m;
problem.r = r;
problem.N = N;

for ii = 1 : N
    values = values_cells{ii};
    indicator = indicator_cells{ii};
    [I, J] = find(indicator == 1);
    values_vec = values(:);
    
    problem.I_cells{ii} = uint32(I);
    problem.J_cells{ii} = uint32(J);
    problem.entries_cells{ii} = values_vec(sub2ind(size(values), I, J));
    problem.nentries_cells{ii} = length(problem.entries_cells{ii});
    problem.masks_cells{ii} = sparse(double(I), double(J), ...
        ones(length(I), 1), size(values, 1), size(values, 2), length(I));
end



% test data
for ii = 1:N
    [m_ii, n_ii] = size(values_cells{ii});
    test_nentries = 2*length(problem.I_cells{ii});
    
    I_test = uint32(randi(m_ii, test_nentries, 1));
    J_test = uint32(randi(n_ii, test_nentries, 1));
    
    YR_ii = YR(1 + (ii-1)*l: min(ii*l, n), :);
    
    entries_test = spmaskmult(YL, YR_ii', uint32(I_test), uint32(J_test));
    
    problem.I_test_cells{ii} = I_test;
    problem.J_test_cells{ii} = J_test;
    problem.entries_test_cells{ii} = entries_test;
    problem.nentries_test_cells{ii} = length(problem.entries_test_cells{ii});
    
end


%% Call decentralized Grassmann
clear options;

options.rho = rho;
options.update = @Grassmann_update;
options.iters = iters;

fprintf('\n---------------------  Decentralized Grassmann ---------------------\n')
[U_cells, infos] = online_decentralized_MC(U_cells0, problem, options);







%% Call decentralized Euclidean
clear options;

options.rho = rho;
options.update = @Euclidean_update;
options.iters = iters;
options.stepsize = 1e0/options.rho; % This works best for this case.
options.stepsize_decay = 10*options.stepsize;% This works best for this case.

% Initialization of U at all the N agents
problem.M = euclideanfactory(m, r); %  Define the Grassmann manifold
for ii = 1 : N
    U_cells0{ii} = problem.M.rand();
end

fprintf('\n---------------------  Decentralized Euclidean ---------------------\n')
[U_cells_euclidean, infos_euclidean] = online_decentralized_MC(U_cells0, problem, options);






%% Plots
pp = 10;

fs = 20;
figure;
semilogy(infos{1}.cost(1:pp:end),'-O','Color','r','LineWidth',2, 'MarkerSize',13);
hold on;
semilogy(infos{2}.cost(1:pp:end),'-s','Color','r','LineWidth',2, 'MarkerSize',13);
semilogy(infos{1}.commcost(1:pp:end),'-+','Color','r','LineWidth',2, 'MarkerSize',13);
semilogy(abs(infos_euclidean{1}.cost(1:pp:end)),':O','Color','b','LineWidth',2, 'MarkerSize',13); % abs() because of numerical inaccuracy.
semilogy(abs(infos_euclidean{2}.cost(1:pp:end)),':s','Color','b','LineWidth',2, 'MarkerSize',13); % abs() because of numerical inaccuracy.
semilogy(abs(infos_euclidean{1}.commcost(1:pp:end)),':+','Color','b','LineWidth',2, 'MarkerSize',13); % abs() because of numerical inaccuracy.
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,['Every ',num2str(pp),'th update'],'FontSize',fs);
ylabel(ax1,'Mean square error on training set','FontSize',fs);
legend('Grassmann agent 1 cost', 'Grassmann agent 2 cost','Grassmann agents 1 - 2 comm. cost','Euclidean agent 1 cost', 'Euclidean agent 2 cost', 'Euclidean agents 1 - 2 comm. cost');
legend 'boxoff';
title([num2str(m),' by ',num2str(n),', rank ',num2str(r),', OS ',num2str(OS)])
box off;



% Test plot when available
if ~isempty(problem.entries_test_cells)
    pp = 10;
    
    fs = 20;
    figure;
    semilogy(infos{1}.cost_test(1:pp:end),'-O','Color','r','LineWidth',2, 'MarkerSize',13);
    hold on;
    semilogy(infos{2}.cost_test(1:pp:end),'-s','Color','r','LineWidth',2, 'MarkerSize',13);
    semilogy(abs(infos_euclidean{1}.cost_test(1:pp:end)),':O','Color','b','LineWidth',2, 'MarkerSize',13); % abs() because of numerical inaccuracy.
    semilogy(abs(infos_euclidean{2}.cost_test(1:pp:end)),':s','Color','b','LineWidth',2, 'MarkerSize',13); % abs() because of numerical inaccuracy.
    hold off;
    ax1 = gca;
    set(ax1,'FontSize',fs);
    xlabel(ax1,['Every ',num2str(pp),'th update'],'FontSize',fs);
    ylabel(ax1,'Mean square error on test set','FontSize',fs);
    legend('Grassmann agent 1 cost', 'Grassmann agent 2 cost','Euclidean agent 1 cost', 'Euclidean agent 2 cost');
    legend 'boxoff';
    title([num2str(m),' by ',num2str(n),', rank ',num2str(r),', OS ',num2str(OS)])
    box off;
    
end

